;   Calc CPM addresses
msize   equ     32                      ; RAM capacity
sectors equ     16                      ; sectors per track for my system
systrks equ     4                       ; my boot system has 4 tracks =-int(-52/16) 
bias    equ     (msize-20-1)*1024       ; adjusted for my system =int((52+4-4*16)/8)
cbase   equ     3400h+bias              ; base of ccp
fbase   equ     cbase+800h              ; base of bdos
bdos    equ     cbase+806h              ; bdos entry
cbios   equ     cbase+1600h             ; base of custom bios (cold boot entry)
spbase  equ     msize*1024              ; base of stack (at top of RAM)
usrdsk  equ     0004h                   ; current user and disk number
iobyte  equ     0003h                   ; intel i/o byte
rst6.5  equ     0034h                   ; interrupt to read serial data input

lf      equ     0ah                     ; line feed
cr      equ     0dh                     ; carriage return

;    Custom CP/M 2.2 BIOS
        org     cbios

;       jump vector for individual subroutines
        jmp     cboot                   ; cold start
wstart:
        jmp     wboot                   ; warm start
        jmp     const                   ; console status
        jmp     conin                   ; console character in
        jmp     conout                  ; console character out
        jmp     list                    ; list character out
        jmp     punch                   ; punch character out
        jmp     reader                  ; reader character out
        jmp     home                    ; move head to home position
        jmp     seldsk                  ; select disk
        jmp     settrk                  ; set track number
        jmp     setsec                  ; set sector number
        jmp     setdma                  ; set dma address
        jmp     read                    ; read disk
        jmp     write                   ; write disk
        jmp     listst                  ; return list status
        jmp     sectran                 ; sector translate

;       System uses a two Flash AT29C256 with 512 x 64b pages (32k)
;       Set as 16 tracks of 16 sectors
;       Assume System uses tracks 0 to 3 (64 sectors)
;       Assume Directory is track 4 (32 entries)
;       Leaving 24kb storage per chip

dpbase:
;       Disk Parameter header for disk 00
        dw      trans,  0000h
        dw      0000h,  0000h
        dw      dirbf,  dpblk32s
        dw      chk00,  all00
;       disk parameter header for disk 01
        dw      trans,  0000h
        dw      0000h,  0000h
        dw      dirbf,  dpblk32d
        dw      chk01,  all01;
;       disk parameter header for disk 02
        dw      trans,  0000h
        dw      0000h,  0000h
        dw      dirbf,  dpblk32d
        dw      chk02,  all02
;       disk parameter header for disk 03
        dw      trans,  0000h
        dw      0000h,  0000h
        dw      dirbf,  dpblk32d
        dw      chk03,  all03

;       sector translate vector (i.e. no translation)
trans:
        db      1,  2,  3,  4           ; sectors  1,  2,  3,  4
        db      5,  6,  7,  8           ; sectors  5,  6,  7,  6
        db      9,  10, 11, 12          ; sectors  9, 10, 11, 12
        db      13, 14, 15, 16          ; sectors 13, 14, 15, 16

;       disk parameter block for system disk (32k)
dpblk32s:    
        dw      sectors                 ; sectors per track
        db      3                       ; block shift factor (block size =1024)
        db      7                       ; block mask
        db      0                       ; null mask
        dw      23                      ; disk size-1 =(32-4*16/128/1024)-1
        dw      31                      ; directory max-1
        db      128                     ; alloc 0 =(31+1)/32 bits
        db      0                       ; alloc 1
        dw      8                       ; check size =(31+1)/4
        dw      systrks                 ; track offset

;       disk parameter block for data disk (32k)
dpblk32d:    
        dw      sectors                 ; sectors per track
        db      3                       ; block shift factor (block size =1024)
        db      7                       ; block mask
        db      0                       ; null mask
        dw      31                      ; disk size-1 =32-1
        dw      31                      ; directory max-1
        db      128                     ; alloc 0 =(31+1)/32 bits
        db      0                       ; alloc 1
        dw      8                       ; check size =(31+1)/4
        dw      0                       ; track offset (no system)

;       BIOS functions

cboot:    
        ; set cold boot defaults only
        di                              ; ? make it obvious
        lxi     sp,spbase               ; use space at top of RAM for stack

        xra     a                       ; zero in the accum
        sta     iobyte                  ; clear the iobyte
        sta     usrdsk                  ; select user and disk zero

        ; display logon message
        mvi     a,0ddh                  ; set SOD high and turn on RST6.5
        sim              
        mvi     a,021h                  ; bitcnt for 9600 baud
        sta     bitcnt
	       lxi	    h,signon                ; Show cold boot message
	       call    prmsg	                  ; print message
        
        jmp     gocpm                   ; initialize and go to cp/m

wboot:
        ; reload CP/M but skip boot on sector 1
        di                              ; ? make it obvious
        lxi     sp,spbase               ; use space at top of RAM for stack

        mvi     a,0dfh                  ; set SOD high and turn off RSTX.X
        sim

        xra     a                       ; select disk 0 (the system disk) and page 0
        out     0ffh                    ; Output to register

        lxi     d,cbase                 ; start of boot dst
        lxi     h,8080h                 ; start of boot src (start at sector 2)
        mvi     c,sectors               ; sectors per track (=16)
        mvi     b,systrks               ; track count (=4)
        push    b                       ; save <bc>
        dcr     c                       ; skip sector 1
wboot$nextsect:
        mvi     b,128                   ; sector size
wboot$copysect:
        mov     a,m
        inx     h
        stax    d
        inx     d
        dcr     b
        jnz     wboot$copysect
        dcr     c
        jnz     wboot$nextsect
        pop     b
        dcr     b
        push    b
        jnz     wboot$nextsect
        jmp     gocpm

;       set parameters and go to cp/m
gocpm:
        mvi     a,0c3h                  ; c3h is a jmp instruction
        sta     0000h                   ; for jmp to wboot
        lxi     h,wstart                ; wboot entry point
        shld    0001h                   ; set address field for jmp at 00000h

        sta     0005h                   ; for jmp to bdos
        lxi     h,bdos                  ; bdos entry point
        shld    0006h                   ; address field of Jump at 0005h to bdos

        sta     rst6.5                  ; for jmp to getchar
        lxi     h,getchar               ; get serial data input entry point
        shld    rst6.5+1                ; set address field of ISR to get serial char

        lxi     b,80h                   ; default dma address is 80h
        call    setdma

        mvi     a,4                     ; define four drives
        sta     drives
        dcr     a                       ; test if usrdsk>=drives
        mov     c,a                     
        lda     usrdsk                  ; get current user and disk number
        cmp     c    
        jc      gocpm$1      
        mvi     a,00h
gocpm$1:
        sta     usrdsk
        mov     c,a                     ; send to the ccp

        mvi     a,0ddh                  ; set SOD high and turn on RSTX.X
        sim
        ei                              ; enable the interrupt system
        jmp     cbase                   ; go to cp/m


;       get serial character and put in the ring buffer (called by RST6.5)
;       use SID for serial input
;       assume a 10 MHz crystal:
;         baud=9600,8,N,1
;         bit count 33 =((10000000/2)/9600-60)/14
getchar:
;       wait for start bit edge
        di                              ; ? timing sensiive
        push    psw
        push    b
        push    d
        push    h
getchar$1:
        rim                             ; 
        ora     a                       ; 
        jm      getchar$1               ; 
;       wait for half bit
        lda     bitcnt                  ; 
        rar                             ;      
        sbi     2                       ; ? adjustment for overheads 
        mov     l,a                     ;
getchar$2:
        dcr     l                       ; 
        jnz     getchar$2               ; 
;       Read in data plus 1 stop bit
        mvi     b,09h                   ; 
getchar$3:
        lda     bitcnt                  ; 13
        mov     l,a                     ; 4
getchar$4:
        dcr     l                       ; 4
        jnz     getchar$4               ; 10/7
        dcr     b                       ; 4
        jz      getchar$5               ; 10   
        rim                             ; 4
        ral                             ; 4
        mov     a,c                     ; 4
        rar                             ; 4
        mov     c,a                     ; 4
        nop                             ; 4    ? match timing
        jnz     getchar$3               ; 10/7 

; Save in ring buffer
getchar$5:
        lda     rput                    ; rbuf put ptr (0-f)
        mov     e,a                     
        lxi     h,rbuf
        mvi     d,00h
        dad     d
        inr     a
        ani     0fh
        sta     rput
        mov     m,c
        pop     h
        pop     d
        pop     b
        pop     psw
        ei
        ret

signon:	                                ; signon message
	db	cr,lf,lf
	db	'my CP/M 2.2 (32k)'
	db	cr,lf,0

; print zero delimited string pointed to by <hl>
prmsg:	
        mov	a,m
        ora	a
        rz                              ; exit on zero
;       more to print
        push h
        mov	c,a
        call putchar
        pop	h
        inx	h
        jmp	prmsg


;       console status, return 0ffh if character ready, 00h if not
const:
        lda     rput
        mov     c,a
        lda     rget
        sub     c
        rz
        mvi     a,0ffh
        ret

;       get console character into A (blocking)
conin:
conin$1:
        lda     rget
        mov     c,a
        lda     rput
        sub     c
        jz      conin$1
        lxi     h,rbuf
        mvi     b,00h
        dad     b
        mov     a,c
        inr     a
        ani     0fh
        sta     rget
        mov     a,m
        ani     7fh
        mov     c,a
        ret


; put character to serial out from register C
; use SOD for serial output
; assumes a 10 MHz crystal:
;   baud=9600,8,N,1
;   bitcnt=33 (~104us)
putchar:
;       set start bit
        xra     a                       ; 
;       loop through bits 
        mvi     b,0ah                   ; 
putchar$1:
        mvi     a,80h                   ; 7
        rar                             ; 4
        sim                             ; 4
;       wait for bit
        lda     bitcnt                  ; 13
        mov     l,a                     ; 4
putchar$2:
        dcr     l                       ; 4
        jnz     putchar$2               ; 10/7
;       output data plus 1 stop bit
        stc                             ; 4
        mov     a,c                     ; 4
        rar                             ; 4
        mov     c,a                     ; 4
        dcr     b                       ; 4
        jnz     putchar$1               ; 10/7
        ret
conout:
        di                              ;
        call    putchar                 ; ? putchar does not alter interrupts
        ei                              ; 
        ret



;       list character from register c
list:
        ret                             ; null subroutine

;       return list status (0 if not ready, 1 if ready)
listst:
        xra    a                        ; 0 is always ok to return
        ret

;       punch character from register C
punch:    
        ret                             ; null subroutine

;       reader character into register a from reader device
reader:
        mvi    a,1ah                    ; enter end of file for now (replace later)
        ani    7fh                      ; remember to strip parity bit
        ret


;       move to the track 0 position of current drive
home:
;       translate this call into a settrk call with parameter 0
        mvi     c,0                     ; select track 0
        mvi     b,0
        call    settrk
        ret                             ; we will move to 00 on first read/write

;       select disk given by register c
seldsk:
        lxi     h,0000h                 ; error return code
        lda     drives
        cmp     c
        rc                              ; out of range
        rz                              ; out of range
        mov     a,c
        sta     diskno
;       select disk selected but page is reset!
        rlc                             ; provision for four disks
        rlc
        ani     0c0h                    ; resets page to 0
        out     0ffh                    ; output to disk/page register
;    compute Disk Parameter Header address
        lda     diskno
        mov     l, a                    ; l=disk number 0, 1, 2, 3
        mvi     h, 0                    ; high order zero
        dad     h
        dad     h
        dad     h
        dad     h
        lxi     d,dpbase
        dad     d                       ; hl=dpbase+diskno*16
        ret
;
settrk:     ; set track given by registers <bc>
        mov     h,b
        mov     l,c
        shld    track
        ret
;
setsec:     ; set sector given by register c
        mov     a, c
        sta     sector
        ret

;       translate the sector <bc> using the translate table [de]
sectran:
        xchg                            ; set hl=translation table
        dad     b                       ; set hl=hl+sector
        mov     l,m                     ; get translated sector
        mvi     h,0                     ; ? sector numbers between 1 and 255
        ret                             ; return with translated sector in <hl>

setdma:
;       set dma address given by registers <bc>
        mov     h,b                     ; high order address
        mov     l,c                     ; low order address
        shld    dmaad                   ; save the address
        ret


;       calculate disk number, disk page, track and sector address
;       returns <cde> = DDTT TTTT 1TTT TSSS S000 0000
calcdskadr:
;       set disk number (0-3)
        lda     diskno                  ; load diskno number
        rrc
        rrc
        ani     0c0h
        mov     c,a                     ; DD00 000
;       set page (first two bits of high byte of track number)
        lda     track+1                 ; load high byte of track number
        rrc
        rrc
        rrc
        rrc
        ani     30h
        ora     c
        mov     c,a                     ; DDTT 0000
;       set high nibble of low byte of track number
        lda     track                   ; load low byte of track number
        rrc
        rrc
        rrc
        rrc
        ani     0fh
        ora     c                       ; DDTT TTTT
        mov     c,a                     ; save

;       set low nibble of sector number
        lda     sector                  ; load sector number
        dcr     a                       ; less one for memory address
        stc
        rar
        mov     d,a                     ; 1SSS SSSS
        mvi     a,00h
        rar                             ; S000 0000
        mov     e,a                     ; save
        mvi     a,87h
        ora     d                       ; 1000 0SSS
;       set low nibble of low byte of track number
        lda     track                   ; load low byte of track number
        rlc
        rlc
        rlc
        ani     78h
        ora     d                       ; 1TTT TSSS
        mov     d,a                     ; save
        ret

;       read sector from disk
read:
        lhld    dmaad                   ; set DMA address <hl>
        call    calcdskadr              ; calculate disk address <cde>
        mov     a,c                     ; out (set) disk number and page
        out     0ffh
        mvi     b,80h                   ; sector size
read$1:
        ldax    d
        inx     d
        mov     m,a
        inx     h
        dcr     b
        jnz     read$1
        xra     a                       ; always succeeds
        ret

;       write sector from disk
write:
        lhld    dmaad                   ; set DMA address <hl>
        call    calcdskadr              ; calculate disk address <cde>
        mov     a,c                     ; out (set) disk number and page
        out     0ffh
        mvi     b,80h                   ; sector size
        xchg                            ; swap destination and source
        mvi     b,40h                   ; page size 64b
write$page1:                            ; write page 1
        ldax    d
        inx     d
        mov     m,a
        inx     h
        dcr     b
        jnz     write$page1
;       wait for flash
        dcx     h
write$check1:
        cmp     m
        jnz     write$check1
        inx     h
        mvi     b,40h                   ; page size 64b
write$page2:                            ; write 2
        ldax    d
        inx     d
        mov     m,a
        inx     h
        dcr     b
        jnz     write$page2
;       wait for flash
        dcx     h
write$check2:
        cmp     m
        jnz     write$check2
        inx     h
        xra     a                       ; always succeeds
        ret

;       BIOS data area
drives: ds      1                       ; number of disk drives 1-16 max
diskno: ds      1                       ; disk number 0-15 max
track:  ds      2                       ; track number 0-65535 max
sector: ds      1                       ; sector number 1-255 max
dmaad:  ds      2                       ; direct memory address


;       scratch ram area for bdos use
dirbf:  ds      128                     ; scratch directory area
all00:  ds      3                       ; allocation vector 0 (=DSM/8+1)
all01:  ds      4                       ; allocation vector 1 (=DSM/8+1)
all02:  ds      3                       ; allocation vector 2 (=DSM/8+1)
all03:  ds      4                       ; allocation vector 3 (=DSM/8+1)
chk00:  ds      8                       ; check vector 0 (if CKS>0, =(DRM+1)/4)
chk01:  ds      8                       ; check vector 1 (if CKS>0, =(DRM+1)/4)
chk02:  ds      8                       ; check vector 2 (if CKS>0, =(DRM+1)/4)
chk03:  ds      8                       ; check vector 3 (if CKS>0, =(DRM+1)/4)

;       Serial use:
rbuf:   ds      16                      ; ring buffer for serial data
rget:   db      0                       ; ring buffer get pointer
rput:   db      0                       ; ring buffer put pointer
bitcnt: db      33                      ; bit count for 9600 baud and 10MHz crystal

        end



