// Include libraries
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdbool.h>

static unsigned char ROM[0x8000];
static int loadI8HEX(char *filename,int offset)
{
  int chr=0;
  int state=0;
  int code=0;
  int count=0;
  int addr=0;
  int type=0;
  int data=0;
  int checksum=0;
  int check=0;
  int addrMin=0xffff;
  int addrMax=0x0000;

  FILE *F1=fopen(filename,"r");
  if (F1==NULL) {
    fprintf(stderr,"%s not found\n",filename);
  } else {
    while ((chr=getc(F1))!=EOF) {
      if (chr==':') {
        code=0;
        state=1;
      } else if ((chr>='0')&&(chr<='9')) {
        code=chr-48;
        state++;
      } else if ((chr>='A')&&(chr<='F')) {
        code=chr-55;
        state++;
      } else if ((chr>='a')&&(chr<='f')) {
        code=chr-87;
        state++;
      } else {
        state=0;
        code=0;
        count=0;
        addr=0;
        type=0;
        data=0;
        checksum=0;
      }

      if (state==2) count=(code&0x0000000f)<<4;
      if (state==3) {
        count+=code&0x0000000f;
        checksum=count;
        count=count*2; // count nibbles
      }
      if (state==4) addr=(code&0x0000000f)<<12;
      if (state==5) addr+=(code&0x0000000f)<<8;
      if (state==6) addr+=(code&0x0000000f)<<4;
      if (state==7) {
        addr+=code&0x0000000f;
        checksum+=addr&0x000000ff;
        checksum+=(addr>>8)&0x000000ff;
      }
      if (state==8) type=(code&0x0000000f)<<4;
      if (state==9) {
        type+=code&0x0000000f;
        checksum+=type;
        if (type==1) state=12;
      }
      if ((state==10)||(state==11)) {
        if (state==10) data=(code&0x0000000f)<<4;
        if (state==11) {
          data+=code&0x0000000f;
          state=9;
          // 32kb RAM
          ROM[(addr+offset)&0x7fff]=(unsigned char)data;
          addr++;
          checksum+=data;
          if (addr>addrMax) addrMax=addr;
          if (addr<addrMin) addrMin=addr;
        }
        count--;
        if (count==0) state=12;
      }
      if (state==13) check=(code&0x0000000f)<<4;
      if (state==14) {
        check+=code&0x0f;
        checksum&=0x0000000ff;
        if (((check+checksum)&0xff)!=0) fprintf(stderr,"Checksum Error\n");
        state=0;
      }
    }
    fclose(F1);
  }
  return addrMax-addrMin;
}

int main(int argc,char **argv) {
  (void) argc;
  (void) argv;
  printf("make CP/M Disk (AlanCooper@StartMail.com)\n");

  // Set director entry
  char filename[13]; // Normal 8.3 file name
  char FileName[12]; // CP/M 8.3 file name
  unsigned short filesize;
  unsigned short offset;
  unsigned short UserCode=0;
  unsigned short ExtCnt;
  unsigned short RecCnt;
  unsigned short BlkCnt;
  unsigned short Extent;
  unsigned short Here;
  unsigned short Disk=0;

  for (Disk=0;Disk<2;Disk++) {
    printf("\nDisk %d:\n",Disk);
    if (Disk==0) {
      // Format Disk
      for (int i=0;i<0x8000;i++) {
        ROM[i&0x7fff]=0x00; // Clear Disk
      }
      for (int i=0;i<32;i++) {
        ROM[(64*128+i*32)&0x7fff]=0xe5; // Flag Exents
      }
      // Load system:
      // start.hex - Jump CBOOT (at 0x0000)
      // ccp.hex   - CCP (at 0x6000)
      // bdos.hex  - BDOS (at 0x6800)
      // cbios.hex - CBIOS (at 0x7600)
      offset=0x8000;
      fprintf(stderr,"  %s %d\n","boot.hex",loadI8HEX("boot.hex",offset));
      offset=0x8000-0x6000+0x0080;
      fprintf(stderr,"  %-16s %8d\n","ccp.hex",loadI8HEX("ccp.hex",offset));
      fprintf(stderr,"  %-16s %8d\n","bdos.hex",loadI8HEX("bdos.hex",offset));
      fprintf(stderr,"  %-16s %8d\n","cbios.hex",loadI8HEX("cbios.hex",offset));

      offset=64*128; // Set start of directory space
    } else {
      // Format Disk
      for (int i=0;i<0x8000;i++) {
        ROM[i&0x7fff]=0x00; // Clear Disk
      }
      for (int i=0;i<32;i++) {
        ROM[(i*32)&0x7fff]=0xe5; // Flag Exents
      }

      offset=0;      // Set start of directory space
    }

    // Set up directory space
    Extent=0xffff;   // Set number of directory entries per file
    Here=1;          // Pointer to next free block (for 32 Extents)
    unsigned short file=0;
    while (true) {
      file++;
      if (Disk==0) {
        if (file==1) strcpy(filename,"stat.hex");
        if (file==2) strcpy(filename,"pip.hex");
        if (file==3) strcpy(filename,"load.hex");
        if (file==4) strcpy(filename,"ddt.hex");
        if (file==5) break;
      } else if (Disk==1) {
        if (file==1) strcpy(filename,"asm.hex");
        if (file==2) strcpy(filename,"ed.hex");
        if (file==3) strcpy(filename,"basic5.hex");
        if (file==4) break;
      } else {
        break;
      }
      filesize=loadI8HEX(filename,offset+Here*1024-256);
      RecCnt=(filesize-1)/128+1;
      BlkCnt=(RecCnt-1)/8+1;
      ExtCnt=BlkCnt/16;
      printf("  offset %d sectors %d\n",offset&0x7fff,((offset&0x7fff)-1)/128+1);
      printf("  filesize %d\n",filesize);
      printf("  RecCnt %d\n",RecCnt);
      printf("  BlkCnt %d\n",BlkCnt);
      printf("  ExtCnt %d\n",ExtCnt);

      for (int i=0;i<11;i++) FileName[i]=' ';FileName[11]=0;
        int i=0;
        while (strchr(" .\0",filename[i])==NULL) {
          if ((filename[i]>='a')&&(filename[i]<='z')) {
            FileName[i]=(char)(filename[i]&0xdf);
          } else {
            FileName[i]=filename[i];
          }
          i++;
          if (i==8) break;
        }
        if (filename[i]=='.') {
          i++;
          int j=8;
          while (strchr(" \0",filename[i])==NULL) {
            if ((filename[i]>='a')&&(filename[i]<='z')) {
              FileName[j]=(char)(filename[i]&0xdf);
          } else {
            FileName[j]=filename[i];
          }
          i++;
          j++;
          if (i==12) break;
        }
      }
      // Change file extension
      FileName[8]='C';
      FileName[9]='O';
      FileName[10]='M';

      bool test=true;
      for (int j=0;j<11;j++) {
        if (strchr(">.,;:=?*[]%|()/\\",FileName[j])!=NULL) test=false;
      }
      if (test) {
        fprintf(stderr,"  Okay %s\n\n",FileName);
      } else {
        fprintf(stderr,"  Fail %s\n\n",FileName);
      }

      // Director Structure:
      //   User code                     1  0
      //   File name                     8  1
      //   Extension                     3  9
      //   Extent number                 1 12 (0-31)
      //   Reserved (0)                  1 13
      //   Reserved (0)                  1 14 (Exend high byte)
      //   record count (1-128)          1 15 (0x80 means Extent is full)
      //   Block addresses              16 16

      // Set allocation table
      int Extension=-1;
      for (int j=0;j<BlkCnt;j++) {
        int k=j%16;
        if (k==0) {
          Extent++;
          Extension++;
          ROM[(offset+32*Extent+0)&0x7fff]=(unsigned char)UserCode;             // User code (0-15)
          strcpy(&ROM[(offset+32*Extent+1)&0x7fff],FileName);                   // File name (8+3)
          ROM[(offset+32*Extent+12)&0x7fff]=(unsigned char)Extension;           // Extension (0-31)
          ROM[(offset+32*Extent+13)&0x7fff]=0;                                  // Reserved (0)
          ROM[(offset+32*Extent+14)&0x7fff]=0;                                  // Reserved (0)
        }
        ROM[(offset+32*Extent+16+k)&0x7fff]=(unsigned char)(Here);              // Set allocation table
        ROM[(offset+32*Extent+15)&0x7fff]=0x80;                                 // Record count (full)
        Here++;
      }
      ROM[(offset+32*Extent+15)&0x7fff]=(unsigned char)(RecCnt%128);            // Update last record Count (partial)
      if (Extent>31) break;
    }

    // Save ROM in I8HEX format
    char DiskName[128];
    sprintf(DiskName,"Disk%d.hex",Disk);
    FILE* F1=fopen(DiskName,"w");
    unsigned char checksum=0;
    for (int i=0;i<0x8000;i++) {
      int j=i%32;
      if (j==0) {
        // printf(":%02X%04X%02X",32,i,0);
        fprintf(F1,":%02X%04X%02X",32,i,0);
        checksum=(unsigned char)(32+((i>>8)+i)+0);
      }
      // printf("%02X",ROM[i&0x7fff]);
      fprintf(F1,"%02X",ROM[i&0x7fff]);
      checksum+=ROM[i&0x7fff];
      if (j==31) {
        // printf("%02X\n",(unsigned char)(-checksum));
        fprintf(F1,"%02X\n",(unsigned char)(-checksum));
      }
    }
    // printf(":00000001FF\n");
    fprintf(F1,":00000001FF\n");
    fclose(F1);

    sprintf(DiskName,"Disk%dA.dat",Disk);
    F1=fopen(DiskName,"w");
    fprintf(F1,"  {\n");
    for (int i=0x0000;i<0x4000;i++) {
      int j=i%16;

      if (j==0) {
        fprintf(F1,"    ");
      }
      if (i%0x4000!=0x3fff) {
        fprintf(F1,"0x%02x,",ROM[i&0x7fff]);
      } else {
        fprintf(F1,"0x%02x",ROM[i&0x7fff]);
      }
      if (j==15) {
        fprintf(F1,"\n");
      }
    }
    fprintf(F1,"  };\n");
    fclose(F1);

    sprintf(DiskName,"Disk%dB.dat",Disk);
    F1=fopen(DiskName,"w");
    fprintf(F1,"  {\n");
    for (int i=0x4000;i<0x8000;i++) {
      int j=i%16;

      if (j==0) {
        fprintf(F1,"    ");
      }
      if (i%0x4000!=0x3fff) {
        fprintf(F1,"0x%02x,",ROM[i&0x7fff]);
      } else {
        fprintf(F1,"0x%02x",ROM[i&0x7fff]);
      }
      if (j==15) {
        fprintf(F1,"\n");
      }
    }
    fprintf(F1,"  };\n");
    fclose(F1);

  }

  return 0;
}

